;; Jess client for WumpusWorld server.
;;
;; This example Jess client uses the basic client-side WW
;; objects to create a connection to the WW server and
;; communicate with it. This is basically a Jess rewrite
;; of the GameClient.java class, using a Javabean class,
;; PerceptKeeperBean, as a wrapper around the communications
;; link.
;;

(import wclient.*)
(import wPercept.TransferPercept)
(import wAction.Action)
(import wworld.ActorState)

;; We use globals to refer to Java objects that we _don't_ want
;; recreated upon reset.
(set-reset-globals FALSE)

;; Grab the constants from the Action class and save them as globals
;; to simplify the Jess code
(defglobal ?*START_GAME* = (get-member Action START_GAME))
(defglobal ?*GOFORWARD* = (get-member Action GOFORWARD))
(defglobal ?*TURN_RIGHT* = (get-member Action TURN_RIGHT))
(defglobal ?*TURN_LEFT* = (get-member Action TURN_LEFT))
(defglobal ?*GRAB* = (get-member Action GRAB))
(defglobal ?*SHOOT* = (get-member Action SHOOT))
(defglobal ?*CLIMB* = (get-member Action CLIMB))
(defglobal ?*SIT* = (get-member Action SIT))
(defglobal ?*NO_OP* = (get-member Action NO_OP))
(defglobal ?*NO_ACTION* = (get-member Action NO_ACTION))
(defglobal ?*END_TRIAL* = (get-member Action END_TRIAL))
(defglobal ?*END_SIMULATION* = (get-member Action END_SIMULATION))

;; Constants from wworld.ActorState
(defglobal ?*ALIVE* = (get-member ActorState ALIVE))
(defglobal ?*DEAD* = (get-member ActorState DEAD))
(defglobal ?*VICTORIOUS* = (get-member ActorState VICTORIOUS))

;; Utility functions
(deffunction updateGUI (?tp)
  "Update the player GUI status display"
  (?*status-frame* updateDisplay ?tp)
  (?*vispercept-frame* makeImage ?tp))

;; Member variables from GameClient.java
(defglobal ?*status-frame* = (new Frame1))      ;; frame in GameClient.java
(defglobal ?*vispercept-frame* = (new Frame2))  ;; frame2 in GameClient.java

;;********************************************************
;;
;; Equivalent to constructor code in GameClient.java
;;
;;********************************************************

;; Create GUI stuff
(?*vispercept-frame* setVisible TRUE)
(?*status-frame* setVisible TRUE)
(?*status-frame* setSize 500 400)
(?*vispercept-frame* setSize 200 300)
(?*status-frame* show)
(?*vispercept-frame* show)

;; Host and port for WumpusWorld server connection
(defglobal ?*hostname* = "localhost")
(defglobal ?*portnum* =  10000)

;; Make the server connection object
(defglobal ?*clientcomm* = (new Communicator ?*hostname* ?*portnum*))

;; Create the percept bean class definition.
(defclass Percepts PerceptKeeperBean)

;;********************************************************
;;
;; Start of player and game state detection rules
;;
;;********************************************************

(defrule detect-start "Detect the start of the game"
  (declare (salience 100))
  (Percepts (actionTaken ?*START_GAME*))
  =>
  (printout t "Start of game" crlf)
  (assert (initial-percept)))

(defrule detect-end "Detect the end of the game"
  (declare (salience 100))
  (Percepts (actionTaken ?*END_SIMULATION*))
  =>
  (printout t "End of simulation" crlf)
  (halt))

(defrule detect-trial-over "Detect end of trial"
  (declare (salience 100))
  (Percepts (actionTaken ?*END_TRIAL*))
  =>
  (printout t "End of trial" crlf))

(defrule check-dead "Detect my death"
  (declare (salience 100))
  (Percepts (status ?*DEAD*))
  =>
  (printout t "I am dead :^(" crlf)
  (halt))

(defrule check-victory "Detect that I've won"
  (declare (salience 100))
  (Percepts (status ?*VICTORIOUS*))
  =>
  (printout t "I am victorious!!" crlf)
  (halt))

;; Let's use this rule to determin that the server is waiting for us
;; to make a real decision. So, rules can use the presence of
;; (response-wait) to trigger the decisionmaking process.
(defrule action-needed "Determine that I need to response to a percept"
  (declare (salience 100))
  (Percepts (status ?*ALIVE*)
	    (actionTaken ~?*END_TRIAL*)
	    (OBJECT ?percept-object))
  =>
  (updateGUI (?percept-object retrievePercept))
  (assert (response-wait)))

;; I don't know if this is necessary, but the example had this.
;; At least, it can serve as an example of a rule that responds to a 
;; Percept with an action
(defrule noop-response "Don't take an action"
  (Percepts (status ~?*ALIVE*)
	    (actionTaken ~?*END_TRIAL*)
	    (OBJECT ?percept-object))
  =>
  (updateGUI (?percept-object retrievePercept))
  (?*clientcomm* writeInt ?*NO_ACTION*))


;; A fixed strategy, from the Java example Player
(deffacts moves "Fixed move strategy"
 (actions ?*GOFORWARD* ?*TURN_LEFT* ?*GOFORWARD* ?*TURN_RIGHT*
		 ?*GOFORWARD* ?*GOFORWARD* ?*GOFORWARD* ?*TURN_LEFT*
		 ?*SHOOT* ?*GOFORWARD* ?*GRAB* ?*TURN_LEFT* ?*GOFORWARD*
		 ?*GOFORWARD* ?*GOFORWARD* ?*TURN_LEFT* ?*GOFORWARD*
		 ?*GOFORWARD* ?*TURN_RIGHT* ?*GOFORWARD* ?*CLIMB*))

(defrule make-a-move "Next move from fixed strategy"
  ?rf <- (response-wait)
  ?af <- (actions ?next-action $?rest)
  =>
  (retract ?rf)
  (retract ?af)
  (assert (actions ?rest))
  (printout t "Performing action " (call Action printAction ?next-action) crlf)
  (?*clientcomm* writeInt ?next-action))

;;********************************************************
;;
;; Startup code
;;
;;********************************************************

; We need reset to ensure all CEs (like not) will work
(reset)

; Make the connection to the server. This could have been done before
; defining the rules, but we might as well wait until after all the
; rules are loaded, just so the clock isn't ticking while Jess is
; loading and interpreting them
(?*clientcomm* createConnection)
(printout t "Created connection; Waiting for server... ")
;; The following statement will block until "start" is clicked on the 
;; server GUI
(defglobal ?*client-id* = (?*clientcomm* readIntFromStream))
(printout t "My ID is " ?*client-id* crlf)
(?*status-frame* setTitle (str-cat "Percepts & Status for AGENT " ?*client-id*))

;; The constructor will start the bean's thread.
(defglobal ?*thePerceptKeeper* = (new PerceptKeeperBean ?*clientcomm*))
(definstance Percepts ?*thePerceptKeeper*)


;; Remember to do:
;; (run-until-halt)

